### PREDATOR PREY SIMULATION
### BY FREDO DURAND
### originally in PYTHON 2.7
### won't work by itself, in particular needs some display code

import numpy.random as rnd

class Animal:
    baseProbaToReproduce=0.0
    def __init__(self, x, y, world):
        self.x=x
        self.y=y
        self.world=world
        world.addAnimal(self)
        self.dead=False
    def simulate(self): pass;
    def reproduce(self, x, y): pass           
    def die(self):
        self.world.setCell(self.x, self.y, None)
        self.dead=True
        self.world.listOfAnimals.remove(self)
    def move(self, x, y):
        self.world.setCell(self.x, self.y, None)
        self.x=x
        self.y=y
        self.world.setCell(x, y, self)
    def moveAndReproduceProba(self, x, y):
        oldX, oldY=self.x, self.y
        self.move(x,y)
        LFriends=self.world.neighborsOfType(self.x, self.y, type(self))
        if rnd.random()<self.baseProbaToReproduce*len(LFriends):
            offspring=self.reproduce(oldX, oldY)        
            
class Prey(Animal):
    baseProbaToReproduce=2.0/8.0
    def reproduce(self, x, y):
        offspring=Prey(x, y, self.world)
    def simulate(self):
        if self.dead: return
        LFree=self.world.freeNeighbors(self.x, self.y)
        LFriends=self.world.neighborsOfType(self.x, self.y, Prey)
        if len(LFree)>0:
            i=rnd.randint(len(LFree))
            self.moveAndReproduceProba(LFree[i][0], LFree[i][1])

class Predator(Animal):
    baseProbaToReproduce=0.1/8.0
    baseProbaToCatchPrey=0.3
    maxTimeWithoutEating=5
    def __init__(self, x, y, world):
        Animal.__init__(self, x, y, world)
        self.timeWithoutEating=0
    def display(self): 
        displayUtilities2.displayBird(self.x*scale, self.y*scale)
    def reproduce(self, x, y):
        offspring=Predator(x, y, self.world)
    def simulate(self):
        if self.dead: print('predator dead:'), self.dead
        if self.dead: return
        self.timeWithoutEating=self.timeWithoutEating+1
        L=self.world.listOfNeighbors(self.x, self.y)
        LFree=self.world.freeNeighbors(self.x, self.y)
        LFriends=self.world.neighborsOfType(self.x, self.y, Predator)
        LPrey=self.world.neighborsOfType(self.x, self.y, Prey)
        if rnd.random()< self.baseProbaToCatchPrey*len(LPrey): #eat a prey   
            i=rnd.randint(len(LPrey))
            prey=self.world.cell(LPrey[i][0], LPrey[i][1])
            prey.die()
            self.moveAndReproduceProba(LPrey[i][0], LPrey[i][1])
            self.timeWithoutEating=0
            return
        if self.timeWithoutEating>self.maxTimeWithoutEating: #starvation
            self.die()
            return
        if len(LFree)>0:            
            i=rnd.randint(len(LFree))
            self.moveAndReproduceProba(LFree[i][0], LFree[i][1])
          
class World:
    def __init__(self, width, height):
        self.width=width
        self.height=height
        self.worldMap=[]
        for y in xrange(height):
            row=[]
            for x in xrange(width):
                row.append(None)
            self.worldMap.append(row)
        self.listOfAnimals=[]
    def cell(self, x, y):
        return self.worldMap[y%self.height][x%self.width]
    def setCell(self, x, y, animal):
        self.worldMap[y%self.height][x%self.width]=animal
    def addAnimal(self, animal):
        c=self.cell(animal.x, animal.y)
        #if c!=None: print 'problem, cell not empty, cannot add ', animal
        self.setCell(animal.x, animal.y, animal)
        self.listOfAnimals.append(animal)
    def listOfNeighbors(self, x, y):
        L=[]
        for j in xrange( y-1, y+2):
            for i in xrange(x-1, x+2):
                L.append((i%self.width, j%self.height))
        L.remove((x, y))
        return L
    def neighborsOfType(self, x, y, myType):
        L=self.listOfNeighbors(x,y)
        L2=[]
        for i, j in L:
            ani=self.cell(i,j)
            if isinstance(ani, myType):
                L2.append((i,j))
        return L2
    def freeNeighbors(self, x, y):
        return self.neighborsOfType(x,y, type(None))
    def runOneIteration(self):
        L=self.listOfAnimals[:] #copy
        for animal in L:
            animal.simulate()

